# flake8: noqa

from mypo.optimizer.minimum_variance_optimizer import MinimumVarianceOptimizer
from mypo.optimizer.optimizer import Optimizer
from mypo.optimizer.sharp_ratio_optimizer import SharpRatioOptimizer



"""Parameters for runner."""
from dataclasses import dataclass

import numpy as np


@dataclass
class Settings:
    """Settings for runner."""

    spending: np.float64
    tax_rate: np.float64
    fee_rate: np.float64



"""Utility functions for model selection."""
from typing import List, Tuple

from mypo.market import Market


def split_n_periods(
    market: Market, n: int, train_span: int = None
) -> Tuple[List[Market], List[Market]]:
    """
    Split market to n periods.

    Parameters
    ----------
    market
        Market data
    n
        Count of split

    train_span
        Train span if you want to specify.

    Returns
    -------
    Split market data
    """
    index = market.get_index()
    len_index = len(index)
    if train_span is None:
        train_span = int(len_index / (n + 1))

    eval_span = int((len_index - train_span) / n)
    train_market: List[Market] = []
    eval_market: List[Market] = []
    for i in range(n):
        start_eval_index = i * eval_span + train_span
        train_market += [
            market.extract(index[start_eval_index - train_span : start_eval_index])
        ]
        eval_market += [
            market.extract(index[start_eval_index : start_eval_index + eval_span])
        ]

    return train_market, eval_market



"""Rebalance strategies."""
import datetime

import numpy as np
import numpy.typing as npt


class Rebalancer(object):
    """Interface class of Rebalance stragegy class."""

    def __init__(self) -> None:
        pass

    def apply(
        self, index: datetime.datetime, assets: npt.ArrayLike, cash: np.float64
    ) -> np.ndarray:
        """
        Apply rebalance strategy to current situation.

        Parameters
        ----------
        index
            Current date for applying rebalance.

        assets
            Current assets for applying rebalance.

        cash
            Current cash for applying rebalance.

        Returns
        -------
        Deal
        """
        pass



"""Rebalance strategies."""
import datetime

import numpy as np
import numpy.typing as npt

from mypo.common import safe_cast
from mypo.rebalancer.rebalancer import Rebalancer


class PlainRebalancer(Rebalancer):
    """Simple weighted rebalance strategy."""

    _weights: np.ndarray

    def __init__(self, weights: npt.ArrayLike) -> None:
        """
        Construct object.

        Parameters
        ----------
        weights
            Weight for applying rebalance.
        """
        super().__init__()
        self._weights = safe_cast(weights)

    def apply(
        self, index: datetime.datetime, assets: npt.ArrayLike, cash: np.float64
    ) -> np.ndarray:
        """
        Apply rebalance strategy to current situation.

        Parameters
        ----------
        index
            Current date for applying rebalance.

        assets
            Current assets for applying rebalance.

        cash
            Current cash for applying rebalance.

        Returns
        -------
        Deal
        """
        diff: np.ndarray = self._weights * np.sum(assets) - safe_cast(assets)
        return diff



"""Class for report result."""
import datetime
from typing import List

import numpy as np
import pandas as pd


class Reporter(object):
    """Reporter class of simulation."""

    _index: List[datetime.datetime]
    _total_assets: List[np.float64]
    _capital_gain: List[np.float64]
    _income_gain: List[np.float64]
    _cash: List[np.float64]
    _deal: List[np.float64]
    _fee: List[np.float64]
    _capital_gain_tax: List[np.float64]
    _income_gain_tax: List[np.float64]

    def __init__(self) -> None:
        """Construct this object."""
        self._index = []
        self._total_assets = []
        self._capital_gain = []
        self._income_gain = []
        self._cash = []
        self._deal = []
        self._fee = []
        self._capital_gain_tax = []
        self._income_gain_tax = []

    def record(
        self,
        index: datetime.datetime,
        total_assets: np.float64,
        capital_gain: np.float64,
        income_gain: np.float64,
        cash: np.float64,
        deal: np.float64,
        fee: np.float64,
        capital_gain_tax: np.float64,
        income_gain_tax: np.float64,
    ) -> None:
        """
        Record current situation.

        Parameters
        ----------
        index
            Current date.

        total_assets
            Current total assets.

        capital_gain
            Current capital gain.

        income_gain
            Current income gain.

        cash
            Current cash.

        deal
            Current deal.

        fee
            Current fee.

        capital_gain_tax
            Current capital gain tax.

        income_gain_tax
            Current income gain tax.
        """
        self._index += [index]
        self._total_assets += [total_assets]
        self._capital_gain += [capital_gain]
        self._income_gain += [income_gain]
        self._cash += [cash]
        self._deal += [deal]
        self._fee += [fee]
        self._capital_gain_tax += [capital_gain_tax]
        self._income_gain_tax += [income_gain_tax]

    def report(self) -> pd.DataFrame:
        """
        Report the result.

        Returns
        -------
        result
        """
        return pd.DataFrame(
            {
                "total_assets": self._total_assets,
                "capital_gain": self._capital_gain,
                "income_gain": self._income_gain,
                "cash": self._cash,
                "deal": self._deal,
                "fee": self._fee,
                "capital_gain_tax": self._capital_gain_tax,
                "income_gain_tax": self._income_gain_tax,
            },
            index=self._index,
        )



"""Optimizer for weights of portfolio."""

import numpy as np
import pandas as pd
from scipy.optimize import minimize

from mypo.common import safe_cast
from mypo.market import Market
from mypo.optimizer.optimizer import Optimizer


class MinimumVarianceOptimizer(Optimizer):
    """Minimum variance optimizer."""

    _historical_data: pd.DataFrame
    _span: int

    def __init__(self, market: Market, span: int = 260):
        """
        Construct this object.

        Parameters
        ----------
        market
            Past market stock prices.

        span
            Span for evaluation.
        """
        self._historical_data = market.get_prices()
        self._span = span

    def optimize_weight(self, minimum_return: float = None) -> np.ndarray:
        """
        Optimize weights.

        Parameters
        ----------
        minimum_return
            Add minimum return constraints.

        Returns
        -------
        Optimized weights
        """
        prices = self._historical_data.tail(n=self._span).to_numpy()
        Q = np.cov(prices.T)
        Q = Q / np.max(np.abs(Q))
        n = Q.shape[0]
        x = np.ones(n) / n

        def fn(x: np.ndarray, Q: np.ndarray) -> np.float64:
            ret: np.float64 = np.dot(np.dot(x, Q), x.T)
            return ret

        cons = [{"type": "eq", "fun": lambda x: np.sum(x) - 1}]
        if minimum_return is not None:
            ret = np.prod(prices, axis=0)
            print(ret)
            cons += [
                {
                    "type": "ineq",
                    "fun": lambda x: np.dot(ret, x) - (1.0 + minimum_return),
                }
            ]

        bounds = [[0.0, 1.0] for i in range(n)]
        minout = minimize(
            fn, x, args=(Q), method="SLSQP", bounds=bounds, constraints=cons
        )
        return safe_cast(minout.x)



"""Simulation."""

import datetime

import numpy as np
import numpy.typing as npt
import pandas as pd

from mypo.common import calc_capital_gain_tax, calc_fee, calc_income_gain_tax, safe_cast
from mypo.market import Market
from mypo.rebalancer.rebalancer import Rebalancer
from mypo.reporter import Reporter
from mypo.settings import Settings

WEEK_DAYS = int(365 * 5 / 7)


class Runner(object):
    """Runner of simulation."""

    _assets: np.ndarray
    _averagel_assets_price: np.ndarray
    _rebalancer: Rebalancer
    _reporter: Reporter
    _cash: np.float64
    _settings: Settings

    def __init__(
        self,
        assets: npt.ArrayLike,
        rebalancer: Rebalancer,
        cash: np.float64,
        settings: Settings,
    ):
        """
        Construct this object.

        Parameters
        ----------
        assets
            Initial asset.

        rebalancer
            Reblance strategy.

        cash
            Initial cash.

        spending
            Monthly spending.
        """
        self._assets = safe_cast(assets)
        self._averagel_assets_price = np.ones(len(self._assets))
        self._rebalancer = rebalancer
        self._reporter = Reporter()
        self._cash = cash
        self._settings = settings
        self._reporter.record(
            pd.NaT,
            self.total_assets(),
            1.0,  # type: ignore
            0.0,  # type: ignore
            self._cash,
            0.0,  # type: ignore
            0.0,  # type: ignore
            0.0,  # type: ignore
            0.0,  # type: ignore
        )

    def total_assets(self) -> np.float64:
        """
        Get current total assets. Total asset is addition of stock assets and cash.

        Returns
        -------
        Total assets.
        """
        return np.float64(np.sum(self._assets) + self._cash)

    def apply(
        self,
        index: datetime.datetime,
        prices: npt.ArrayLike,
        price_dividends_yield: npt.ArrayLike,
        expense_ratio: npt.ArrayLike,
    ) -> None:
        """
        Apply current market situation.

        Parameters
        ----------
        index
            Current date.

        prices
            Current market situation.

        price_dividends_yield
            Current dividends yield this date.

        expense_ratio
            Expense ratio of holding assets.
        """
        previous_assets = np.sum(self._assets)
        prices = safe_cast(prices)
        price_dividends_yield = safe_cast(price_dividends_yield)
        expense_ratio = safe_cast(expense_ratio)

        # apply market prices
        self._assets = self._assets * prices
        diff = self._rebalancer.apply(index, self._assets, self._cash)
        deal: np.float64 = np.abs(diff)

        # process of capital gain
        capital_gain_tax = calc_capital_gain_tax(
            self._averagel_assets_price, self._assets, diff, self._settings.tax_rate
        )
        self._cash -= capital_gain_tax
        fee = calc_fee(diff, self._settings.fee_rate)
        self._cash -= fee
        self._cash -= np.float64(np.sum(diff))
        self._assets += diff
        trading_prices = np.where(diff > 0, prices, self._averagel_assets_price)
        self._averagel_assets_price = (
            self._averagel_assets_price * previous_assets + diff * trading_prices
        ) / self._assets

        # process of income gain
        income_gain = np.sum(self._assets * price_dividends_yield)
        income_gain_tax = calc_income_gain_tax(
            self._assets, price_dividends_yield, self._settings.tax_rate
        )
        self._cash += income_gain
        self._cash -= income_gain_tax

        # process of others
        self._assets = (1.0 - expense_ratio / WEEK_DAYS) * self._assets

        # record to reporter
        capital_gain: np.float64 = np.float64(
            np.sum(self._assets) / np.sum(previous_assets)
        )
        self._reporter.record(
            index,
            self.total_assets(),
            capital_gain,
            income_gain,
            self._cash,
            deal,
            fee,
            capital_gain_tax,
            income_gain_tax,
        )

    def run(self, market: Market) -> None:
        """
        Run simulation.

        Parameters
        ----------
        market
            Market data.
        """
        index = market.get_index()
        markets = market.get_prices().to_records(index=False)
        price_dividends_yield = market.get_price_dividends_yield().to_records(
            index=False
        )
        expense_ratio = market.get_expense_ratio()
        for i in range(len(markets)):
            self.apply(index[i], markets[i], price_dividends_yield[i], expense_ratio)

    def report(self) -> pd.DataFrame:
        """
        Report simulation.

        Returns
        -------
        result
        """
        return self._reporter.report()



"""Market object for store and loading stock prices data."""

from __future__ import annotations

import pickle
from typing import Dict

import numpy as np
import pandas as pd

from mypo.common import safe_cast


class Market(object):
    """Market class for store and loading stock prices data."""

    _tickers: Dict[str, pd.DataFrame]
    _expense_ratio: Dict[str, float]

    def __init__(
        self, tickers: Dict[str, pd.DataFrame], expense_ratio: Dict[str, float]
    ):
        self._tickers = tickers
        self._expense_ratio = expense_ratio

    def save(self, filepath: str) -> None:
        """
        Save market data to file.

        Parameters
        ----------
        filepath
            Path to file for storing data.
        """
        with open(filepath, "wb") as bin_file:
            pickle.dump(self, bin_file)

    @classmethod
    def load(cls, filepath: str) -> Market:
        """
        Load market data from file.

        Parameters
        ----------
        filepath
            Path to file for loading data.

        Returns
        -------
        Market object

        """
        with open(filepath, "rb") as bin_file:
            value: Market = pickle.load(bin_file)
            return value

    def extract(self, index: pd.Series) -> Market:
        """
        Extract market data.

        Parameters
        ----------
        index
            Index what you want to extract.

        Returns
        -------
            Extracted Data
        """
        return Market(
            tickers={ticker: data.loc[index] for ticker, data in self._tickers.items()},
            expense_ratio=self._expense_ratio,
        )

    def get_index(self) -> pd.Series:
        """
        Get index date from stored market data.

        Returns
        -------
        index date
        """
        rs = [self._tickers[ticker][["Close"]] for ticker in self._tickers.keys()]
        df = pd.concat(rs, axis=1, join="inner")
        return df.index

    def get_prices(self) -> pd.DataFrame:
        """
        Get price data from stored market data.

        Returns
        -------
        Price data
        """
        rs = [self._tickers[ticker][["Close"]] for ticker in self._tickers.keys()]
        df = pd.concat(rs, axis=1, join="inner")
        for c in df.columns:
            df[c] = df[c].pct_change() + 1.0
        df.fillna(1.0, inplace=True)
        df.columns = self._tickers.keys()
        return df

    def get_price_dividends_yield(self) -> pd.DataFrame:
        """
        Get price dividends yield from stored market data.

        Returns
        -------
        price dividends yield data
        """
        rs = [
            self._tickers[ticker]["Dividends"] / self._tickers[ticker]["Close"]
            for ticker in self._tickers.keys()
        ]
        df = pd.concat(rs, axis=1, join="inner")
        df.columns = self._tickers.keys()
        return df

    def get_expense_ratio(self) -> np.ndarray:
        """
        Get expense ratio.

        Returns
        -------
        Expense ratio
        """
        rs = [self._expense_ratio[ticker] for ticker in self._tickers.keys()]
        return safe_cast(rs)



"""
Utility functions.

A collection of utlities by using in Runner class.

"""


from typing import Any

import numpy as np
import numpy.typing as npt


def safe_cast(value: Any) -> np.ndarray:
    """
    Cast array like type to numpy array.

    Parameters
    ----------
    value
        target value

    Returns
    -------
    casted_value
    """
    casted_value = np.array(list(value))
    return casted_value


def calc_capital_gain_tax(
    average_asset_prices: npt.ArrayLike,
    prices: npt.ArrayLike,
    diff: npt.ArrayLike,
    tax_rate: np.float64,
) -> np.float64:
    """
    Calculate capital gain from market data and assets.

    Parameters
    ----------
    average_asset_prices
        average assets from investment started.

    prices
        Current assets.

    diff
        Deal.

    tax_rate
        Ratio of tax for capital gain.

    Returns
    -------
    capital_gain_tax
    """
    average_asset_prices = safe_cast(average_asset_prices)
    prices = safe_cast(prices)
    diff = safe_cast(diff)
    deal_sell = np.where(diff > 0, 0, diff)
    unrealized_gain = prices - average_asset_prices
    plus_unrealized_gain = np.where(unrealized_gain > 0, unrealized_gain, 0)
    capital_gain_tax: np.float64 = np.sum(plus_unrealized_gain * deal_sell * tax_rate)
    return capital_gain_tax


def calc_income_gain_tax(
    assets: npt.ArrayLike, price_dividends_yield: npt.ArrayLike, tax_rate: np.float64
) -> np.float64:
    """
    Calculate income gain from market data and assets.

    Parameters
    ----------
    assets
        Current assets.

    price_dividends_yield
        Price dividends yield.

    tax_rate
        Ratio of tax for capital gain.

    Returns
    -------
    income_gain_tax
    """
    assets = safe_cast(assets)
    price_dividends_yield = safe_cast(price_dividends_yield)
    income_gain_tax: np.float64 = -np.sum(assets * price_dividends_yield * tax_rate)
    return income_gain_tax


def calc_fee(diff: npt.ArrayLike, fee_rate: np.float64) -> np.float64:
    """
    Calculate fee for dealing.

    Parameters
    ----------
    diff
        Deal.

    fee_rate
        Ratio of fee for trading.

    Returns
    -------
    fee
    """
    diff = safe_cast(diff)
    deal = np.abs(diff)
    fee: np.float64 = -np.sum(deal * fee_rate)
    return fee



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mypo', 'mypo.loss_function', 'mypo.optimizer', 'mypo.rebalancer']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.20.1,<2.0.0',
 'pandas>=1.2.2,<2.0.0',
 'pymc3>=3.11.1,<4.0.0',
 'scipy>=1.6.0,<2.0.0',
 'yfinance>=0.1.55,<0.2.0']

setup_kwargs = {
    'name': 'mypo',
    'version': '0.0.39',
    'description': '',
    'long_description': '# mypo\n\ndocumatation is [here](https://sonesuke.github.io/mypo/).',
    'author': 'sonesuke',
    'author_email': 'iamsonesuke@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sonesuke/mypo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9.1,<4.0.0',
}


setup(**setup_kwargs)



# flake8: noqa
__version__ = "0.0.39"

from .common import calc_capital_gain_tax, calc_fee, calc_income_gain_tax
from .loader import Loader
from .market import Market
from .model_selection import split_n_periods
from .reporter import Reporter
from .runner import Runner
from .settings import Settings



"""Rebalance strategies."""
import datetime

import numpy as np
import numpy.typing as npt

from mypo.common import safe_cast
from mypo.rebalancer.rebalancer import Rebalancer


class MonthlyRebalancer(Rebalancer):
    """Weighted rebalance strategy by monthly applying."""

    _old_month: int
    _weights: np.ndarray

    def __init__(self, weights: npt.ArrayLike, old_month: int = 0) -> None:
        """
        Construct object.

        Parameters
        ----------
        weights
            Weight for applying rebalance.

        old_month
            Previous month.
        """
        super().__init__()
        self._old_month = old_month
        self._weights = safe_cast(weights)

    def apply(
        self, index: datetime.datetime, assets: npt.ArrayLike, cash: np.float64
    ) -> np.ndarray:
        """
        Apply rebalance strategy to current situation.

        Parameters
        ----------
        index
            Current date for applying rebalance.

        assets
            Current assets for applying rebalance.

        cash
            Current cash for applying rebalance.

        Returns
        -------
        Deal
        """
        assets = safe_cast(assets)
        if self._old_month != index.month:
            self._old_month = index.month
            diff: np.ndarray = self._weights * np.sum(assets) - assets
            return diff
        else:
            zero: np.ndarray = assets - assets
            return zero



# flake8: noqa

from mypo.rebalancer.monthly_rebalancer import MonthlyRebalancer
from mypo.rebalancer.plain_rebalancer import PlainRebalancer
from mypo.rebalancer.rebalancer import Rebalancer
from mypo.rebalancer.threshold_rebalancer import ThresholdRebalancer
